﻿using System;
using SharpDX;
using SharpDX.Direct2D1;
using SharpDX.Direct3D10;
using SharpDX.Direct3D9;
using D2DFactory = SharpDX.Direct2D1.Factory;
using Device1 = SharpDX.Direct3D10.Device1;
using DWriteFactory = SharpDX.DirectWrite.Factory;
using Format = SharpDX.Direct3D9.Format;
using QueryType = SharpDX.Direct3D9.QueryType;
using Surface = SharpDX.Direct3D9.Surface;
using Usage = SharpDX.Direct3D9.Usage;

namespace MiniTri
{

  public class SharedTextureContainer : IDisposable
  {
    protected Texture _texture;
    protected Surface _surface;
    protected Texture2D _texture10;
    protected IntPtr _dxgiResourceSharedHandle = IntPtr.Zero;
    protected DeviceEx _device9;
    protected Device1 _device10;
    protected D2DFactory _factory;
    protected int _width;
    protected int _height;
    protected AbstractDeviceSync _sync9;
    protected AbstractDeviceSync _sync10;

    public RenderTarget RenderTarget2D { get; private set; }

    public Texture SharedTexture9
    {
      get { return _texture; }
    }

    public Surface SharedSurface9
    {
      get { return _surface; }
    }

    public Texture2D SharedTexture10
    {
      get { return _texture10; }
    }

    public DeviceEx Device9
    {
      get { return _device9; }
    }

    public Device1 Device10
    {
      get { return _device10; }
    }

    public D2DFactory D2DFactory
    {
      get { return _factory; }
    }

    public delegate void Render9Delegate(DeviceEx device);

    public delegate void Render10Delegate(RenderTarget renderTarget);

    public SharedTextureContainer(DeviceEx device9, Device1 device10, int width, int height)
    {
      _device9 = device9;
      _device10 = device10;
      _width = width;
      _height = height;
      _sync9 = new Direct3D9Sync(device9);
      _sync10 = new Direct3D10Sync(device10);
    }

    public void Allocate()
    {
      _texture = new Texture(_device9,
        _width,
        _height,
        1,
        Usage.RenderTarget,
        Format.A8R8G8B8, // equivalent to DXGI.Format.R8G8B8A8_UNorm
        Pool.Default,
        ref _dxgiResourceSharedHandle);

      _surface = _texture.GetSurfaceLevel(0);

      _texture10 = _device10.OpenSharedResource<Texture2D>(_dxgiResourceSharedHandle);

      _factory = new D2DFactory();

      using (var surface10 = _texture10.QueryInterface<SharpDX.DXGI.Surface>())
      {
        RenderTarget2D = new RenderTarget(_factory, surface10,
          new RenderTargetProperties(new PixelFormat(SharpDX.DXGI.Format.B8G8R8A8_UNorm, AlphaMode.Premultiplied)));
      }
      RenderTarget2D.AntialiasMode = AntialiasMode.PerPrimitive;
    }

    public void Render9(Render9Delegate action)
    {
      _sync9.Begin();
      var oldTarget = _device9.GetRenderTarget(0);
      _device9.SetRenderTarget(0, _surface);
      action(_device9);
      _device9.SetRenderTarget(0, oldTarget);
      _sync9.End();
      _sync9.WaitForFinished();
    }

    public void Render10(Render10Delegate action)
    {
      _sync10.Begin();
      action(RenderTarget2D);
      _sync10.End();
      _sync10.WaitForFinished();
    }

    public void Dispose()
    {
      if (_surface != null)
        _surface.Dispose();
      if (_texture != null)
        _texture.Dispose();
      if (_texture10 != null)
        _texture10.Dispose();
      if (_factory != null)
        _factory.Dispose();
      if (_sync9 != null)
        _sync9.Dispose();
      if (_sync10 != null)
        _sync10.Dispose();
    }
  }

  public abstract class AbstractDeviceSync : IDisposable
  {
    public abstract void Dispose();
    public abstract void Begin();
    public abstract void End();
    public abstract void WaitForFinished();
  }

  public class Direct3D9Sync : AbstractDeviceSync
  {
    protected SharpDX.Direct3D9.Query _query;

    public Direct3D9Sync(DeviceEx device)
    {
      _query = new SharpDX.Direct3D9.Query(device, QueryType.Event);
    }

    public override void Dispose()
    {
      _query.Dispose();
    }

    public override void Begin()
    {
      _query.Issue(Issue.Begin);
    }

    public override void End()
    {
      _query.Issue(Issue.End);
    }

    public override void WaitForFinished()
    {
      bool data;
      while (!_query.GetData(out data, true))
      {
        // No operation, will be kind of busy wait? Thread.Sleep will introduce at least 1 ms delay
      }
    }
  }

  public class Direct3D10Sync : AbstractDeviceSync
  {
    protected SharpDX.Direct3D10.Query _query;

    public Direct3D10Sync(Device1 device)
    {
      QueryDescription qd = new QueryDescription {Flags = QueryFlags.None, Type = SharpDX.Direct3D10.QueryType.Event};
      _query = new SharpDX.Direct3D10.Query(device, qd);
    }

    public override void Dispose()
    {
      _query.Dispose();
    }

    public override void Begin()
    {
      _query.Begin();
    }

    public override void End()
    {
      _query.End();
    }

    public override void WaitForFinished()
    {
      bool finished = false;
      while (!finished)
      {
        using (DataStream dataStream = _query.GetData())
          finished = dataStream.ReadBoolean();
      }
    }
  }
}